module Metap

open FStar.Tactics

let int42 : int = _ by (exact (`42))

let suc (x:int) = x + 1

let rec repeatN (n:nat) (t : unit -> Tac unit) : Tac unit =
  if n = 0
  then ()
  else (t (); repeatN (n-1) t)

let ten : int = _ by (repeatN 10 (fun () -> apply (`suc)); exact (`0))

(* Print the definition of `ten`: it's not `10`! *)

let rec sums (n:nat) : Tac unit =
  if n = 0
  then exact (`1)
  else begin
    apply (`(+));
    sums (n-1);
    sums (n-1)
  end

let big : int = _ by (sums 4)

(*
We can also use the reflection capabilities of F* to inspect the
goals we need to solve and act accordingly. For instance, here is a
metaprogram that constructs an addition function for any amount of
arguments.
*)

let rec __run (bs:binders) : Tac unit =
  match inspect (cur_goal ()) with
  | Tv_Arrow _ _ ->
    let b = intro () in
    __run (b::bs)
  | _ ->
    iter (fun b -> apply (`(+));
                exact (binder_to_term b)) bs;
    exact (`0)

let make_add () : Tac unit =
  __run []

let add_2 : a:int -> b:int -> Tot int = _ by (make_add ())
let add_3 : a:int -> b:int -> c:int -> Tot int = _ by (make_add ())
let add_4 : a:int -> b:int -> c:int -> d:int -> Tot int = _ by (make_add ())

let _ = assert (add_2 3 4 == 7)
let _ = assert (add_3 3 4 5 == 12)
let _ = assert (add_4 3 4 5 6 == 18)

(******* Meta args *******)

let diag (x : int) (#[exact (quote x)] y : int) : tuple2 int int = (x, y)

let _ = assert (diag 5 == (5, 5))

let _ = assert (diag 5 #23 == (5, 23))

open FStar.Tactics.Typeclasses

class add 'a = {
  sum : 'a -> 'a -> 'a
}

(* {|add a|} is just sugar for `(#[tcresolve ()] _ : add a)` *)
let add3 (#a:Type) {|add a|} (x y z : a) : a = sum x (sum y z)

(* Instances are just values with a `tcinstance` attribute, which is
what `tcresolve` looks for. *)
instance add_int : add int = { sum = (+) }

(* Same as
[@tcinstance] let add_int : add int = { sum = (+) }
*)

let test_add = assert (add3 1 2 3 == 6)
